<?php
namespace Teknisa\GOP\Services;

use PHPExcel_IOFactory;
use Zeedhi\Framework\DTO;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DB\StoredProcedure\Param;
use Zeedhi\Framework\DB\StoredProcedure\StoredProcedure;

class Gop10000SheetImporter 
{
    const SHEETS_FOLDER = __DIR__ . "/../Temp/";
    const SHEET_TEMPLATE = __DIR__ . "/../Temp/Modelo.xlsx";
    
    const COLUMNS_MAP = array(
        "CDFILIAL" => 
            array(
                "index"     => 0,
                "maxLength" => 4,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Filial"
            ),
        "NRANO" => 
            array(
                "index"     => 1,
                "maxLength" => 4,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Ano"
            ),
        "CDITMORCA" => 
            array(
                "index"     => 2,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Código Orçamento"
            ),
        "NMORCA" => 
            array(
                "index"     => 3,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "Nome Orçamento"
            ),
        "CDDESPDIV" => 
            array(
                "index"     => 4,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "Código"
            ),
        "NM" => 
            array(
                "index"     => 5,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "Nome"
            ),
        "CDARVPROD" => 
            array(
                "index"     => 6,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "Código Produto"
            ),
        "NMPRODUTO" => 
            array(
                "index"     => 7,
                "nullable"  => true,
                "zerofill"  => false,
                "label"     => "Nome Produto"
            ),
        "VRORCAJAN" => 
            array(
                "index"     => 8,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Janeiro"
            ),
        "VRORCAFEV" => 
            array(
                "index"     => 9,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Fevereiro"
            ),
        "VRORCAMAR" => 
            array(
                "index"     => 10,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Março"
            ),
        "VRORCAABR" => 
            array(
                "index"     => 11,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Abril"
            ),
        "VRORCAMAI" => 
            array(
                "index"     => 12,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Maio"
            ),
        "VRORCAJUN" => 
            array(
                "index"     => 13,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Junho"
            ),
        "VRORCAJUL" => 
            array(
                "index"     => 14,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Julho"
            ),
        "VRORCAAGO" => 
            array(
                "index"     => 15,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Agosto"
            ),
        "VRORCASET" => 
            array(
                "index"     => 16,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Setembro"
            ),
        "VRORCAOUT" => 
            array(
                "index"     => 17,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Outubro"
            ),
        "VRORCANOV" => 
            array(
                "index"     => 18,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Novembro"
            ),
        "VRORCADEZ" => 
            array(
                "index"     => 19,
                "nullable"  => false,
                "zerofill"  => false,
                "label"     => "Dezembro"
            )
    );
    
    protected $errors = array();
    protected $errorMessages = array();
    protected $newCodeSC;

    public function __construct($connection, $environment, $newCodeSC) 
    {
        $this->connection = $connection;
        $this->environment = $environment;
        $this->newCodeSC = $newCodeSC;
    }
    
    private function generateNewCode($tableName) {
        return $this->newCodeSC->getCodeByTableName($tableName, $this->environment->getNrOrg());
    }
    
    public function createTemporaryFile($sheetImport) 
    {
        if ( isset($sheetImport) &&  is_array($sheetImport) &&  !empty($sheetImport) 
             && isset($sheetImport[0]["b64File"]) && isset($sheetImport[0]["name"]) ) {
            $sheetB64 = $sheetImport[0]["b64File"];
            $sheetName = $sheetImport[0]["name"];
            $sheet = $this->decodeFromBase64($sheetB64);
            $filePath = self::SHEETS_FOLDER . "tmp-" . $sheetName;
            file_put_contents($filePath, $sheet);
            chmod($filePath, 0644);
            return $filePath;
        } else {
            throw new \Exception ("sheetUploadFailure");
        }
    }
    
    private function decodeFromBase64($sheet) 
    {
        $b64File = explode(",", $sheet);
        $base64_decode = base64_decode($b64File[1], true);

        return $base64_decode;
    }
    
    public function loadSheet($sheetPath) 
    {
        try {
            $objPHPExcel = PHPExcel_IOFactory::load($sheetPath);
            return $objPHPExcel->getActiveSheet();
        } catch( \Excention $e) {
            throw new \Exception("invalidSheet");
            
        }
    }
    
    public function parseSheet($worksheet) 
    {
        try {
            $rows = array();
            
            foreach ($worksheet->getRowIterator() AS $row) {
                $cellIterator = $row->getCellIterator();
                $cellIterator->resetEnd("T");
                $cellIterator->setIterateOnlyExistingCells(FALSE); // This loops through all cells,
                $cells = [];
                $hasNotNullCell = false;
                
                foreach ($cellIterator as $cell) {
                    $value = $cell->getValue();
                    $hasNotNullCell = $hasNotNullCell || $value;
                    $cells[] = $value;
                }
                
                if ($hasNotNullCell) { // At least one cell with value
                    $rows[] = $cells;
                }
            }
            
            if (((is_array($rows) && !is_null($rows)) ? count($rows) : 0) < 2) {
                throw new \Exception("emptySheet");
            }
        
            return $rows;
        } catch( \Excention $e) {
            throw $e;
        }
    }
    
    public function validateSheetColumns($sheet) 
    {
        if (isset($sheet[0])) {
            $header = $sheet[0];
            
            $count = (is_array($header) && !is_null($header)) ? count($header) : 0;

            if ($count == 20) {
                foreach (self::COLUMNS_MAP as $column) {
                    if ($header[$column["index"]] != $column["label"]) {
                        throw new \Exception("invalidSheetColumnsLabels");
                    }
                }
            }
        } else {
            throw new \Exception("invalidSheet");
        }
    }
    
    public function validateSheetRows(&$sheet, $response) 
    {
        foreach ($sheet as $key => &$row) {
            if ($key != 0) { // Skip Header Row
                $this->validateEmptyValues($row, $key);
                $this->validateInvalidValues($row, $key);
                
            }
        }
        
        if (!empty($this->errors)) {
            $this->decodeErrorMessages();
            $response->addDataSet(new DataSet("errors", $this->errorMessages));
            throw new \Exception("sheetErrors", 1);
        }
    }
    
    private function validateEmptyValues(&$row, $rowNum) 
    {
        foreach (self::COLUMNS_MAP as $columnName => $column) {
            if (is_null($row[$column["index"]]) && !$column["nullable"]) {
                $this->addError($rowNum, "emptyCell", $columnName);
            }
        }
    }
    
    private function validateInvalidValues(&$row, $rowNum) 
    {
        foreach (self::COLUMNS_MAP as $columnName => $column) {
            if (strlen($row[$column["index"]]) != 0) {
                if (isset($column["length"]) && strlen($row[$column["index"]]) != $column["length"]) {
                    if (strlen($row[$column["index"]]) < $column["length"] && isset($column["zerofill"]) && $column["zerofill"]) {
                        $row[$column["index"]] = str_pad($row[$column["index"]], $column["length"], "0", STR_PAD_LEFT);
                    } else {
                        $this->addError($rowNum, "invalidLength", $columnName); 
                    }
                } else if (isset($column["maxLength"]) && strlen($row[$column["index"]]) > $column["maxLength"]) {
                    $this->addError($rowNum, "invalidMaxLength", $columnName); 
                }
            }
        }
    }
    
    private function decodeErrorMessages() 
    {
        $tab = "          ";
        
        foreach ($this->errors as $rowNum => $error) {
            $message = "Linha " . ($rowNum + 1) . ":";
            array_push($this->errorMessages, $message);
            
            if (isset($error["emptyCell"])) {
                foreach (($error["emptyCell"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" deve ser informado.";
                    array_push($this->errorMessages, $message);
                }
            }

            if (isset($error["invalidLength"])) {
                foreach (($error["invalidLength"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" deve possuir exatamente ". self::COLUMNS_MAP[$column]["length"] . " caracteres.";
                    array_push($this->errorMessages, $message);
                }
            }

            if (isset($error["invalidMaxLength"])) {
                foreach (($error["invalidMaxLength"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" deve possuir no máximo ". self::COLUMNS_MAP[$column]["maxLength"] . " caracteres.";
                    array_push($this->errorMessages, $message);
                }
            }

            if (isset($error["unitNotFound"])) {
                foreach (($error["unitNotFound"]) as $column) {
                    $message = $tab . "A \"" . self::COLUMNS_MAP[$column]["label"] . "\" informada não foi encontrada. Esperado -> Código com 4 dígitos de uma unidade exitente no sistema.";
                    array_push($this->errorMessages, $message);
                }
            }
            
            if (isset($error["productNotFound"])) {
                foreach (($error["productNotFound"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" informado não foi encontrado. Esperado -> Código de um produto exitente no sistema.";
                    array_push($this->errorMessages, $message);
                }
            }
            
            if (isset($error["dateInvalid"])) {
                foreach (($error["dateInvalid"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" informado não é um ano válido. Esperado -> Ano válido com 4 dígitos";
                    array_push($this->errorMessages, $message);
                }
            }
            
            if (isset($error["numberInvalid"])) {
                foreach (($error["numberInvalid"]) as $column) {
                    $message = $tab . "O \"" . self::COLUMNS_MAP[$column]["label"] . "\" informado não é um número válido. Esperado -> Apenas números e positivo.";
                    array_push($this->errorMessages, $message);
                }
            }
        }
    }
    
    private function addError($rowNum, $errorType, $error) 
    {
        if (!isset($this->errors[$rowNum])) {
            $this->errors[$rowNum] = array();
        }
        
        if (!isset($this->errors[$rowNum][$errorType])) {
            $this->errors[$rowNum][$errorType] = array();
        }
        
        array_push($this->errors[$rowNum][$errorType], $error);
    }
    
    public function validateAndInsertData($row, $sheet, $response)
    {
       
        $dataArrays = self::getDataArray($sheet);
        self::validateImportedData($dataArrays, $response);
        self::insertImportedData($dataArrays, $row);
        return $dataArrays;
    }
    
    private function getDataArray($sheet)
    {
        $dataArrays = [];
        foreach ($sheet as $ind => $data) {
            if($ind != 0){
                array_push($dataArrays, $data);
            }
        }
        return $dataArrays;
    }
    
    private function validateImportedData($dataArrays, $response){
        foreach ($dataArrays as $line => $dataArray) {
            foreach($dataArray as $key => $data) {
                switch ($key) {
                    case 0:
                        $unit = $this->verifyUnitExists($data);
                        
                        if (empty($unit)) {
                            $this->addError($line, "unitNotFound", "CDFILIAL");
                        }
                    break;
                    
                    case 1:
                        $year = false;
                        
                        if (strlen($data) == "4") {
                            $year = \DateTime::createFromFormat("Y", $data);
                        }
                        
                        if (!$year || $year->format("Y") != $data) {
                            $this->addError($line, "dateInvalid", "NRANO");
                        }
                    break;
                    
                    case 2:
                    case 4:
                    case 8:
                    case 9:
                    case 10:
                    case 11:
                    case 12:
                    case 13:
                    case 14:
                    case 15:
                    case 16:
                    case 17:
                    case 18:
                    case 19:
                        if ($data != null) {
                            if (is_numeric($data)) {
                                if ($data < 0) {
                                    $this->addError($line, "numberInvalid", array_keys(self::COLUMNS_MAP)[$key]);
                                }
                            } else {
                                $this->addError($line, "numberInvalid", array_keys(self::COLUMNS_MAP)[$key]);
                            }
                        }
                    break;
                    
                    case 6:
                        $product = $this->verifyProductExists($data);
                        
                        if (empty($unit)) {
                            $this->addError($line, "productNotFound", "CDARVPROD");
                        }
                    break;
                }
            }
        }
        
        if (!empty($this->errors)) {
            $this->decodeErrorMessages();
            $response->addDataSet(new DataSet("errors", $this->errorMessages));
            throw new \Exception("sheetErrors", 1);
        }
    }
    
    private function getInsertQuery()
    {
        $sql = "INSERT INTO INTGORCAGOP
                (
                    NRINTGORCAGOP,
                    CDFILIAL,
                    NRANO,
                    CDITMORCA,
                    CDDESPDIV,
                    CDARVPROD,
                    VRORCAJAN,
                    VRORCAFEV,
                    VRORCAMAR,
                    VRORCAABR,
                    VRORCAMAI,
                    VRORCAJUN,
                    VRORCAJUL,
                    VRORCAAGO,
                    VRORCASET,
                    VRORCAOUT,
                    VRORCANOV,
                    VRORCADEZ,
                    IDORCAIMP
                )
                VALUES
                (
                    :P_NRINTGORCAGOP,
                    :P_CDFILIAL,
                    :P_NRANO,
                    :P_CDITMORCA,
                    :P_CDDESPDIV,
                    :P_CDARVPROD,
                    :P_VRORCAJAN,
                    :P_VRORCAFEV,
                    :P_VRORCAMAR,
                    :P_VRORCAABR,
                    :P_VRORCAMAI,
                    :P_VRORCAJUN,
                    :P_VRORCAJUL,
                    :P_VRORCAAGO,
                    :P_VRORCASET,
                    :P_VRORCAOUT,
                    :P_VRORCANOV,
                    :P_VRORCADEZ,
                    :P_IDORCAIMP
                )";
                
        return $sql;
    }
    
    private function insertImportedData($dataArrays)
    {
        $sql = $this->getInsertQuery();

        $newCode = (int)$this->generateNewCode("INTGORCAGOP");
        
        foreach ($dataArrays as $dataArray) {
            $date = "01/01/$dataArray[1]";
            
            $param = array();
            $param["P_NRINTGORCAGOP"] = $newCode;
            $param["P_CDFILIAL"] = (string)$dataArray[0];
            $param["P_NRANO"] = $date;
            $param["P_CDITMORCA"] = (int)$dataArray[2];
            $param["P_CDDESPDIV"] = $dataArray[4];
            $param["P_CDARVPROD"] = (string)$dataArray[6];
            $param["P_VRORCAJAN"] = (float)$dataArray[8];
            $param["P_VRORCAFEV"] = (float)$dataArray[9];
            $param["P_VRORCAMAR"] = (float)$dataArray[10];
            $param["P_VRORCAABR"] = (float)$dataArray[11];
            $param["P_VRORCAMAI"] = (float)$dataArray[12];
            $param["P_VRORCAJUN"] = (float)$dataArray[13];
            $param["P_VRORCAJUL"] = (float)$dataArray[14];
            $param["P_VRORCAAGO"] = (float)$dataArray[15];
            $param["P_VRORCASET"] = (float)$dataArray[16];
            $param["P_VRORCAOUT"] = (float)$dataArray[17];
            $param["P_VRORCANOV"] = (float)$dataArray[18];
            $param["P_VRORCADEZ"] = (float)$dataArray[19];
            $param["P_IDORCAIMP"] = "N";
            try {
                $std = $this->connection->prepare($sql);
                $std->execute($param);
            } catch (Exception $e){
                throw new \Exception("invalidSheet");
            }
        }
    }
    
    private function verifyUnitExists($data)
    {
        $sql = "SELECT
                    CDFILIAL
                FROM
                    FILIAL
                WHERE
                    CDFILIAL = :P_CDFILIAL";
                    
        $param = array();
        $param["P_CDFILIAL"] = $data;
        
        try {
            $std = $this->connection->prepare($sql);
            $std->execute($param);
            return $std->fetchAll();
        } catch (Exception $e){
            throw new \Exception("invalidSheet");
        }
    }
    
    private function verifyProductExists($data)
    {
        $sql = "SELECT
                    CDARVPROD
                FROM
                    PRODUTO
                WHERE
                    CDARVPROD = :P_CDARVPROD";
                    
        $param = array();
        $param["P_CDARVPROD"] = $data;
        
        try {
            $std = $this->connection->prepare($sql);
            $std->execute($param);
            return $std->fetchAll();
        } catch (Exception $e){
            throw new \Exception("invalidSheet");
        }
    }
    
    public function removeTemporaryFile($filePath) 
    {
        unlink($filePath);
    }
    
    public function startImportProcedure()
    {
        $cdoperador = $this->environment->getCdOperador();
        
        $procedure = new StoredProcedure($this->connection, 'IMPORTAORCAGOP');
        $procedure->addParam(new Param('P_CDOPERADOR', Param::PARAM_INPUT, $cdoperador, Param::PARAM_TYPE_STR));
        
        $procedure->execute();
    }
    
    public function checkForErrors($dataArrays)
    {
        foreach ($dataArrays as $line => $dataArray) {
            foreach($dataArray as $key => $value) {
                switch ($key) {
                    case 0:
                        $listaFili = '_' . $value;
                    break;
                    case 1:
                        $listaAno = '_' . $value;
                    break;
                }
            }
        }
        
        $sql = "SELECT
                    CDFILIAL,
                    DSMOTIVO
                FROM
                    LOGPREPARAORCAGOP
                WHERE INSTR(:P_LISTAFILI, CDFILIAL) > 0 
                AND INSTR(:P_LISTAANO, NRANOORCAGOP) > 0";
        
        $param = array(
            'P_LISTAFILI' => $listaFili,
            'P_LISTAANO'  => $listaAno
        );
        
        
        $std = $this->connection->prepare($sql);
        $std->execute($param);
        return $std->fetchAll();
    }
}